using System;
using System.Collections.Generic;

namespace InterpolationTest
{
    public class ClientSim
    {
        public bool isExtrapolating = false;
        public double renderTime = 0.0;
        public double clientTime = 0.0;
        public bool firstSnapshot = true;
        public bool smoothTimerState = true;

        private double physicsLastExecuted = 0.0;
        private double renderLastExecuted = 0.0;

        public Vector2D localPos1 = new Vector2D(0, 0);
        public Vector2D localVel1 = Settings.defaultVelocity;
        public Vector2D renderPos1 = new Vector2D(0, 0);
        public Vector2D renderVel1 = Settings.defaultVelocity;
        public Vector2D localPos2 = new Vector2D(0, 0);
        public Vector2D localVel2 = Settings.defaultVelocity;
        public Vector2D renderPos2 = new Vector2D(0, 0);
        public Vector2D renderVel2 = Settings.defaultVelocity;

        public List<Snapshot> snapshots = new List<Snapshot>();
        public double[] measuredLatency = new double[Settings.latencySampleAmmount];
        public int measuredLatencyIndex;
        public double targetTime;

        public DateTime targetSetTime;
        private System.Diagnostics.Stopwatch swex = new System.Diagnostics.Stopwatch();
        private System.Diagnostics.Stopwatch swup = new System.Diagnostics.Stopwatch();

        private int findSnapshot(double t)
        {
            for (int i = 0; i < snapshots.Count - 1; ++i)
            {
                Snapshot snapshot0 = snapshots[i + 0];
                Snapshot snapshot1 = snapshots[i + 1];
                if (t >= snapshot0.m_time && t <= snapshot1.m_time)
                {
                    return i;
                }
            }

            return -1;
        }

        public void addSnapshot(double snapTime, Vector2D pos, Vector2D vel)
        {
            if (firstSnapshot)
            {
                firstSnapshot = false;

                renderTime = snapTime - Settings.timeOffset;
                targetTime = renderTime;
                localPos2 = pos;
                localVel2 = vel;
            }

            Snapshot snapshot = new Snapshot();
            snapshot.m_time = snapTime;
            snapshot.m_pos = pos;
            snapshot.m_vel = vel;
            snapshots.Add(snapshot);

            setTargetTime(snapTime);
        }

        public void setTargetTime(double snapTime)
        {
            measuredLatency[measuredLatencyIndex] = clientTime - snapTime;

            if (measuredLatencyIndex == Settings.latencySampleAmmount - 1)
            {
                measuredLatencyIndex = 0;
            }
            else
            {
                ++measuredLatencyIndex;
            }

            double maxLatency = 0;
            double sumLatency = 0;
            for (int i = 0; i < Settings.latencySampleAmmount; ++i)
            {
                maxLatency = Math.Max(maxLatency, Math.Abs(measuredLatency[i]));
                sumLatency += measuredLatency[i];
            }

            double avrLatency = sumLatency / Settings.latencySampleAmmount;
            double instantTargetTime = clientTime - avrLatency - Settings.timeOffset;
            double instantdifferenceTime = Math.Abs(renderTime - instantTargetTime);
            double differenceTime = Math.Abs(renderTime - targetTime);

            if (swex.Elapsed.TotalMilliseconds / 1000 >= Settings.extrapoTime || instantdifferenceTime > Settings.reactDifference)
            {
                targetTime = instantTargetTime;
                System.Diagnostics.Debug.WriteLine("setTargetTime: SET at clientTIme: " + clientTime);
            }

            //SetTarget Debug - Activate to view it.
            //System.Diagnostics.Debug.WriteLine("avrLatency: " + avrLatency);
            //System.Diagnostics.Debug.WriteLine("instantdifferenceTime: " + instantdifferenceTime);
            //System.Diagnostics.Debug.WriteLine("differenceTime: " + differenceTime);
        }

        // Lerp that is framerate-independent.See https://www.scirra.com/blog/ashley/17/using-lerp-with-delta-time for the explanation and maths !
        public double lerpDT(double a, double b, double f, double dt)
        {
            double t = 1.0 - Math.Pow(f, dt);
            return ((a) * (1 - (t)) + (b) * (t));
        }

        public void update(double elapsed)
        {
            //System.Diagnostics.Debug.WriteLine("elapsedTime: " + elapsed);

            clientTime += elapsed;
            renderTime += elapsed;
            targetTime += elapsed;

            
            // simulate physics - Lomsor Experimental
            if (clientTime >= physicsLastExecuted + Settings.physicsStepClient)
            {
                localPos1 = localPos1 + localVel1 * (clientTime - physicsLastExecuted); // move player 1 (that one is local only, no interpolation)
                localPos2 = localPos2 + localVel2 * (clientTime - physicsLastExecuted); // move player 2 (that's the one server-interpolated)

                physicsLastExecuted = clientTime;
                System.Diagnostics.Debug.WriteLine("Client Physics Exececuted at clientTime: " + clientTime);
            }
            
            /*
            // simulate physics - Original
            localPos1 = localPos1 + localVel1 * elapsed; // move player 1 (that one is local only, no interpolation)
            localPos2 = localPos2 + localVel2 * elapsed; // move player 2 (that's the one server-interpolated)
            */

            if (smoothTimerState)
            {
                // speed up or slow down the display time to match the target time
                // that's pretty much the cause of the stuttering, however without that code, changes in latency mean the client and server timers are no longer synchronized properly
                // ie. try to PageUp to increase the latency to 1 second -> the client extrapolates all the time (green dot filled)
                double delta = Math.Abs(renderTime - targetTime) * 1.0;
                delta = Math.Max(delta, 0.0);
                delta = Math.Min(delta, 1.0);
                //System.Diagnostics.Debug.WriteLine("delta: " + delta);
                renderTime = lerpDT(renderTime, targetTime, delta, elapsed);
                //displayTime = targetTime;
            }
         
            // process local snapshots
            while (snapshots.Count > 0)
            {
                Snapshot first = snapshots[0];
                if (first.m_time < renderTime - Settings.snapStoreTime)
                {
                    // discard too old snapshots
                    snapshots.RemoveAt(0);
                }
                else
                {
                    break;
                }
            }


            // Render and Snapshot usage - Lomsor Experimental
            if (clientTime >= renderLastExecuted + Settings.renderStep)
            {
                //Render Local Player
                renderPos1 = localPos1;

                //Render Network Targets
                int currentSnapshotIDrender = findSnapshot(renderTime);
                if (currentSnapshotIDrender == -1)
                {
                    // extrapolate latest snapshot
                    isExtrapolating = true;

                    swex.Start();

                    if (snapshots.Count > 0)
                    {
                        Snapshot lastSnapshot = snapshots[snapshots.Count - 1];

                        double t = renderTime - lastSnapshot.m_time;
                        renderPos2 = lastSnapshot.m_pos + lastSnapshot.m_vel * t;
                    }
                }
                else
                {
                    isExtrapolating = false;

                    swex.Reset();

                    //System.Diagnostics.Debug.WriteLine("snapshots[currentSnapshotIDrender].m_time: " + snapshots[currentSnapshotIDrender].m_time);

                    Snapshot snapshot0 = snapshots[currentSnapshotIDrender + 0];
                    Snapshot snapshot1 = snapshots[currentSnapshotIDrender + 1];


                    // interpolate between current and next snapshot See this: https://gamedev.stackexchange.com/questions/12754/how-to-interpolate-between-two-game-states

                    double elapsedSinceUpdate = renderTime - snapshot0.m_time;

                    double tr = elapsedSinceUpdate / Settings.snapStep;
                    tr = Math.Max(tr, 0.0);
                    tr = Math.Min(tr, 1.0);

                    //System.Diagnostics.Debug.WriteLine("renderTime: " + renderTime);
                    //System.Diagnostics.Debug.WriteLine("snapshot0.m_time: " + snapshot0.m_time);
                    //System.Diagnostics.Debug.WriteLine("snapshot1.m_time: " + snapshot1.m_time);
                    //System.Diagnostics.Debug.WriteLine("elapsedSinceUpdate: " + elapsedSinceUpdate);
                    //System.Diagnostics.Debug.WriteLine("snapInterval: " + Settings.snapInterval);
                    //System.Diagnostics.Debug.WriteLine("tr: " + tr);

                    renderPos2 = snapshot0.m_pos + (snapshot1.m_pos - snapshot0.m_pos) * tr;
                    renderVel2 = snapshot0.m_vel + (snapshot1.m_vel - snapshot0.m_vel) * tr;
                    //renderPos2 = snapshot0.m_pos;
                    //renderVel2 = snapshot0.m_vel;
                }
                renderLastExecuted = clientTime;
                System.Diagnostics.Debug.WriteLine("Client Render Exececuted at clientTime: " + clientTime);
            }
        }
    }
}
